/*
 *  acm : an aerial combat simulator for X
 *  Copyright (C) 1991-1998  Riley Rainey
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 dated June, 1991.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program;  if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave., Cambridge, MA 02139, USA.
 */

#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "adf.h"
#include "aps.h"
#include "box.h"
#include "browse.h"
#include "commands.h"
#include "dis_if.h"
#include "drone.h"
#include "flaps.h"
#include "gear.h"
#include "hud.h"
#include "hsi.h"
#include "instruments.h"
#include "joystick.h"
#include "mouse.h"
#include "navaid.h"
#include "planes.h"
#include "players.h"
#include "pm.h"
#include "prompt.h"
#include "radar.h"
#include "render.h"
#include "terminal.h"
#include "update.h"
#include "weapon.h"
#include "windows.h"

#define MAX_MAPPED_STRING_LEN	20
#define MAX_POPUP_STRING_LEN	40

#define FLIGHT_STATE_FILE "./acm-flight-state"

/*
	To prevent unwanted quits from the program, it is requested to
	press SHIFT-Q two times within 2 seconds.  Anyway, the user can
	still close the window. Here we store the last time the SHIFT-Q
	was pressed:
*/
static double pressed_XK_Q_time = -5.0;


static void
doButtonEvent(craft * c, viewer * u, XEvent * ev)
{
	if ( ! u->hasFocus | ! u->hasComm )  return;

	if (ev->xbutton.button == Button1) {
		selectCockpitItem( c, u, 
						   ev->xbutton.x, 
						   ev->xbutton.y, 
						   ev->xbutton.time );
	}
	else if (ev->xbutton.button == Button2) {
		fireWeapon(c);
	}
	else if (ev->xbutton.button == Button3) {
		selectWeapon(c);
	}
}


static void
doButtonReleaseEvent(craft * c, viewer * u, XEvent * ev, int player)
{

	if (ev->xbutton.button == Button2)
		ceaseFireWeapon(c);
}


#define SWITCH_TO_NEXT_PLAYER_ENABLED

#ifdef SWITCH_TO_NEXT_PLAYER_ENABLED

static void
switchToNextPlayer(viewer * v)
{
	int    i;
	craft *p;

	for ((i = 0, p = ptbl); i < MAXPLAYERS; (++i, ++p)) {
		if ( ((p->type == CT_DRONE) || (p->type == CT_PLANE))
		&& p->vl == NULL ) {
			break;
		}
	}

	if ( i == MAXPLAYERS ) {
		prompt_viewer_print(v, "No other plane available");
		return;
	}

	v->c->vl = NULL; /* FIXME: vl might contain several viewers */
	v->c = p;
	p->vl = v;
	/****
	if ( p->type == CT_DRONE )
		v->hasComm = FALSE;
	else
		v->hasComm = TRUE;
	****/
}

#endif


static _BOOL
doKeyEvent(craft * c, viewer * u, XEvent * ev)
/*
	Handle key press event. Return TRUE is player gets killed or player
	asked to terminate the program.
*/
{

	KeySym    keysym;
	XComposeStatus compose;
	char      buffer[MAX_MAPPED_STRING_LEN];
	int       buflen = MAX_MAPPED_STRING_LEN;

#ifdef SPECIAL_KEYS
	FILE     *fp;
	craft     pentry;

#endif

	if ( ! u->hasFocus ) {
		return FALSE;
	}

	(void) XLookupString((XKeyEvent *) ev, buffer, buflen,
						 &keysym, &compose);

	if ( ! u->hasComm ){
		if ( keysym == XK_d )
			u->hasComm = TRUE;
		return FALSE;
	}

	if ( terminal_enabled(u) && (keysym != XK_F1) ){
		terminal_input_char(u, buffer[0]);
		{
			char line[100];
			if( terminal_read_string(u, line, 100) )
				commands_execute(u, line);
		}
		return FALSE;
	}

	switch (keysym) {

	case XK_F1:
		terminal_toggle(u);
		break;

#ifdef SWITCH_TO_NEXT_PLAYER_ENABLED
	case XK_Tab:
		switchToNextPlayer(u);
		break;
#endif

	case XK_d:
		u->hasComm = FALSE;
		break;

	case XK_H:
		windows_set_layout(c, u, u->w->width, u->w->height, !u->hud_mode);
		break;

	case XK_A:
		aps_ap_toggle(c);
		break;

	case XK_Z:
		aps_ap_toggle(c);
		if( aps_ap_enabled(c) )
			aps_ap_set_vs(c, 0.0);
		break;

	case XK_X:
		aps_ac_toggle(c);
		break;

	case XK_T:
		aps_at_toggle(c);
		break;

	case XK_N:
		aps_an_toggle(c);
		break;

	case XK_L:
		aps_al_toggle(c);
		break;

	case XK_parenleft:
		aps_bank_max_dec(c);
		break;

	case XK_parenright:
		aps_bank_max_inc(c);
		break;

	case XK_M:
		/* Toggle MH<->TH display */
		c->showMag = ! c->showMag;
		break;

	case XK_less:
		aps_an_disable(c);
		aps_al_disable(c);
		aps_aw_set(c, DEGtoRAD(-3.0));
		break;

	case XK_greater:
		aps_an_disable(c);
		aps_al_disable(c);
		aps_aw_set(c, DEGtoRAD(3.0));
		break;

	case XK_comma:
		aps_an_disable(c);
		aps_al_disable(c);
		aps_aw_set(c, DEGtoRAD(-1.5));
		break;

	case XK_period:
		aps_an_disable(c);
		aps_al_disable(c);
		aps_aw_set(c, DEGtoRAD(1.5));
		break;

	case XK_bar:
		aps_an_disable(c);
		aps_al_disable(c);
		aps_aw_set(c, DEGtoRAD(0.0));
		break;

	case XK_slash:
		aps_an_disable(c);
		aps_al_disable(c);
		aps_aw_disable(c);
		break;

#ifdef sun
	case XK_R2:
#else
	case XK_Up:
#endif
		c->pitchComm = c->pitchComm + 0.01;
		if ( c->pitchComm > 1.0 )  c->pitchComm = 1.0;
		break;

#ifdef sun
	case XK_R5:
#else
	case XK_Down:
#endif
		c->pitchComm = c->pitchComm - 0.01;
		if ( c->pitchComm < -1.0 )  c->pitchComm = -1.0;
		break;

	case XK_z:
		c->rudderComm = c->rudderComm + 0.01;
		if ( c->rudderComm > 1.0 )  c->rudderComm = 1.0;
		break;

	case XK_c:
		c->rudderComm = c->rudderComm - 0.01;
		if ( c->rudderComm < -1.0 )  c->rudderComm = -1.0;
		break;

	case XK_x:
		c->rudderComm = 0.0;
		break;

#ifdef sun
	case XK_R7:
#else
	case XK_Home:
#endif
		aps_off(c);
		c->SeTrim = 0.0;
		c->SaTrim = 0.0;
		break;

	case XK_j:
		/* Adjust Elevator Trim */
		c->SeTrim -= c->pitchComm;
		break;

	case XK_y:
		flaps_up(c);
		break;

	case XK_h:
		flaps_down(c);
		break;

	case XK_w:
		flaps_speed_brakes_retract(c);
		break;

	case XK_s:
		flaps_speed_brakes_extend(c);
		break;

	case XK_t:
		if( u->hud_mode )
			hud_timer_toggle(u);
		else
			inst_timer_toggle(u);
		break;

	case XK_1:
	case XK_Next:

		aps_al_disable(c);
		aps_an_disable(c);
		aps_ap_disable(c);
		aps_at_disable(c);
		aps_aw_disable(c);

		c->throttleComm = 8192;
		break;

	case XK_2:
	case XK_KP_Subtract:
		if ( aps_at_enabled(c) ) {
			aps_at_dec_velocity(c);
		}
		else {
			c->throttleComm -= 512;
			if ( c->throttleComm < 6553 )
				c->throttleComm = 6553;
		}
		break;

	case XK_3:
	case XK_KP_Add:
		if ( aps_at_enabled(c) ) {
			aps_at_inc_velocity(c);
		}
		else {
			c->throttleComm += 512;
			if ( c->throttleComm >= 32768 )
				c->throttleComm = 32768;
		}
		break;

	case XK_4:
	case XK_Prior:

		aps_al_disable(c);
		aps_an_disable(c);
		aps_at_disable(c);
		aps_ap_disable(c);
		aps_aw_disable(c);

		c->throttleComm = 32768;
		break;
	
	case XK_exclam:
		pm_thrust_reverse_toggle(c);
		break;

	case XK_a:
		pm_after_burner_toggle(c);
		break;

	case XK_space:
		/* Switch NAVn/RNAVn receivers: */
		aps_al_disable(c);
		aps_an_disable(c);
		hsi_switch_mode(u);
		break;

	case XK_7:
		aps_an_disable(c);
		hsi_obs_inc(u, -1);
		adf_hdg_inc(u, -1);
		break;

	case XK_8:
		aps_an_disable(c);
		hsi_obs_inc(u, +1);
		adf_hdg_inc(u, +1);
		break;

	case XK_9:
		aps_al_disable(c);
		aps_an_disable(c);
		hsi_frq_inc(u, -1);
		adf_frq_inc(u, -1);
		break;

	case XK_0:
		aps_al_disable(c);
		aps_an_disable(c);
		hsi_frq_inc(u, +1);
		adf_frq_inc(u, +1);
		break;

	case XK_F3:
		hsi_theta_inc(u, -1);
		break;

	case XK_F4:
		hsi_theta_inc(u, +1);
		break;

	case XK_F5:
		hsi_rho_inc(u, -1);
		break;

	case XK_F6:
		hsi_rho_inc(u, +1);
		break;

	case XK_b:
		c->flags ^= FL_BRAKES;
		break;

	case XK_g:
		gear_handle_toggle(c);
		break;

	case XK_l:
		drone_new(c, NULL);
		break;

#ifdef sun
	case XK_Up:
#else
	case XK_KP_8:
#endif
		setOutsideView(c, u, VIEW_FORWARD);
		prompt_viewer_print(u, "Front view");
		break;

/* look right */

#ifdef sun
	case XK_Right:
#else
	case XK_KP_6:
#endif
		setOutsideView(c, u, VIEW_RIGHT);
		prompt_viewer_print(u, "Right view");
		break;

/* look left */

#ifdef sun
	case XK_Left:
#else
	case XK_KP_4:
#endif
		setOutsideView(c, u, VIEW_LEFT);
		prompt_viewer_print(u, "Left view");
		break;

/* look back */

#ifdef sun
	case XK_Down:
#else
	case XK_KP_2:
#endif
		setOutsideView(c, u, VIEW_AFT);
		prompt_viewer_print(u, "Rear view");
		break;

/* look up */

#ifdef sun
	case XK_R11:
#else
	case XK_KP_5:
#endif
		setOutsideView(c, u, VIEW_UP);
		prompt_viewer_print(u, "Up view");
		break;

	case XK_KP_0:
		setOutsideView(c, u, VIEW_DOWN);
		prompt_viewer_print(u, "Down view");
		break;

	case XK_n:
		if( c->flags & FL_CHASE_VIEW ){
			setOutsideView(c, u, VIEW_FORWARD);
			prompt_viewer_print(u, "Front view");

		} else if( c->cinfo->object == NULL ){
			prompt_viewer_print(u, "Sorry, no object model available for this plane.");

		} else {
			setOutsideView(c, u, VIEW_CHASE);
			prompt_viewer_print(u, "Chase view");

		}
		break;

	case XK_q:
		c->curRadarTarget = newRadarTarget(c);
#ifdef HAVE_DIS
		dis_radarTargetChanged(c);
#endif
		break;

	case XK_D:
		/* Debug */
		FSPageToggle();
		break;

	case XK_r:
		aps_al_disable(c);
		aps_an_disable(c);

		switch (c->radarMode) {

		case RM_OFF:
		case RM_STANDBY:
			if( u->c->cinfo->radarOutput > 0.0 ){
				/* we really have a radar -- enable it */
				setRadarMode(c, RM_NORMAL);
				enableRadar(c, RM_NORMAL);
			} else {
				/* no radar -- HSI mode */
				disableRadar(c);
				setRadarMode(c, RM_HSI);
				hsi_enable(u);
			}
			break;

		case RM_NORMAL:
		case RM_ACM:
		case RM_STT:
			disableRadar(c);
			setRadarMode(c, RM_HSI);
			hsi_enable(u);
			break;

		case RM_HSI:
			hsi_disable(u);
			setRadarMode(c, RM_ADF);
			adf_enable(u);
			break;

		case RM_ADF:
			adf_disable(u);
			setRadarMode(c, RM_OFF);
			break;

		default:
			adf_disable(u);
			disableRadar(c);
			setRadarMode(c, RM_OFF);
			hsi_enable(u);
			break;
		}

		break;

#ifdef HAVE_DIS
/* this is a hack to allow change of beam parameter index
in the outgoing electromagnetic emission pdu's */
	case XK_R:
		aps_al_disable(c);
		aps_an_disable(c);
		switch (c->radarMode) {
		case RM_OFF:
		case RM_STANDBY:
			setRadarMode(c, RM_NORMAL);
			enableRadar(c, RM_NORMAL);
			break;
		case RM_NORMAL:
			disableRadar(c);
			setRadarMode(c, RM_ACM);
			enableRadar(c, RM_ACM);
			break;
		case RM_ACM:
			disableRadar(c);
			setRadarMode(c, RM_STT);
			enableRadar(c, RM_STT);
			break;
		case RM_STT:
			disableRadar(c);
			setRadarMode(c, RM_HSI);
			hsi_enable(u);
			break;
		case RM_HSI:
			disableRadar(c);
			hsi_disable(u);
			setRadarMode(c, RM_OFF);
			break;
		default:
			abort();
		}
		break;
#endif

#ifdef FIXME_INCOMPLETE_OR_BUGGED

	case XK_i:
		fp = fopen(FLIGHT_STATE_FILE, "w");
		if ( fp == NULL ) {
			char s[1000];

			sprintf("Error opening the file " FLIGHT_STATE_FILE
				" for writing: %s", strerror(errno));
			prompt_craft_print(c, s);
		}
		else {
			fwrite((char *) &ptbl[c->pIndex],
				sizeof(craft), 1, fp);
			fclose(fp);
			prompt_craft_print(c, "Flight state saved in file "
				FLIGHT_STATE_FILE);
		}
		break;

	case XK_I:
		fp = fopen(FLIGHT_STATE_FILE, "r");
		if ( fp == NULL ) {
			char s[1000];

			sprintf("Error opening the file " FLIGHT_STATE_FILE
				" for reading: %s", strerror(errno));
			prompt_craft_print(c, s);
		}
		else {
			fread((char *) &pentry, sizeof(craft), 1, fp);
			fclose(fp);

			c->flags = pentry.flags;
			c->Sg = pentry.Sg;
			c->prevSg = pentry.prevSg;
			c->Cg = pentry.Cg;
			c->trihedral = pentry.trihedral;
			c->curHeading = pentry.curHeading;
			c->curPitch = pentry.curPitch;
			c->curRoll = pentry.curRoll;
			c->p = pentry.p;
			c->q = pentry.q;
			c->r = pentry.r;
			c->damageBits = pentry.damageBits;
			c->structurePts = pentry.structurePts;
			c->damageCL = pentry.damageCL;
			c->damageCM = pentry.damageCM;
			c->curThrust = pentry.curThrust;
			c->throttle = pentry.throttle;

			/* FIXME: there are missing fields */

			prompt_craft_print(c, "Flight state restored from file "
				FLIGHT_STATE_FILE);
		}
		break;

	case XK_semicolon:
		debug ^= 1;
		break;
#endif /* FIXME_INCOMPLETE_OR_BUGGED */

	//case XK_Escape:
	case XK_Q:
		if ( pressed_XK_Q_time + 2.0 >= curTime ){
			killPlayer(c, NULL);
			return TRUE;
		}
		else {
			pressed_XK_Q_time = curTime;
			prompt_viewer_print(u, "Press SHIFT-Q again within 2 sec. to exit the program");
		}
		break;

	case XK_P:
		system("xwd -name " ACM " -out /tmp/acm-dump-`date +%s`");
		prompt_viewer_print(u, "Screen dumped in file /tmp/acm-dump-*");
		break;

#ifdef FIXME_INCOMPLETE_OR_BUGGED

	case XK_braceleft:
		startBlackBoxRecording();
		prompt_viewer_print(u, "Black box recording started");
		break;

	case XK_braceright:
		endBlackBoxRecording();
		prompt_viewer_print(u, "Black box recording stopped");
		break;

	case XK_bracketleft:
		startBlackBoxPlayback();
		prompt_viewer_print(u, "Black box playback started");
		break;

#endif

	case XK_K:
		CalibrateJoystick();
		break;

	case XK_F7:
		inst_altimeter_correction(u, -1);
		break;

	case XK_F8:
		inst_altimeter_correction(u, +1);
		break;

	case XK_F9:
		inst_attitude_reset(u);
		break;

	case XK_F11:
		inst_attitude_adjust_pitch(u, -DEGtoRAD(0.5));
		break;

	case XK_F12:
		inst_attitude_adjust_pitch(u, DEGtoRAD(0.5));
		break;

	case XK_plus:
		windows_zoom_in(u);
		break;

	case XK_minus:
		windows_zoom_out(u);
		break;

	case XK_dollar:
		if( c->type == CT_PLANE )
			drone_take_commands(c);
		else
			drone_release_commands(c);
		break;

	}

	return FALSE;
}


static _BOOL
doWindowEvent(viewer * u)
/*
	Return TRUE on:
	
	- close request of the window: in this case the viewer u gets closed
	  and released
	
	- player gets killed or terminated: in this case the viewer u and all
	  the other viewers attached to the craft c gets closed and the craft c
	  gets released
*/
{
	XEvent    ev;
	craft    *c;

	c = u->c;

	if (XCheckWindowEvent(u->dpy, u->win, -1L, &ev)) {

		switch (ev.type) {

		case MappingNotify:
			XRefreshKeyboardMapping(&ev.xmapping);
			break;

		case KeyPress:
			if (u->viewer_state == ViewerStateBrowsing ||
				u->viewer_state == ViewerStatePiggyback) {
				if (doBrowseKeyEvent( c, u, &ev, 1)) {
					return FALSE;
				}
			}
			else {
				if( doKeyEvent(c, u, &ev) ){
					return TRUE;
				}
			}
			break;

		case ButtonPress:
			doButtonEvent(c, u, &ev);
			break;

		case ButtonRelease:
			doButtonReleaseEvent(c, u, &ev, 1);
			break;

		case ConfigureNotify:
			windows_set_layout(c, u,
				ev.xconfigure.width,
				ev.xconfigure.height, u->hud_mode);
			break;

		case FocusIn:
			u->hasFocus = TRUE;
			break;

		case FocusOut:
			u->hasFocus = FALSE;
			u->hasComm = FALSE;
			break;

		case Expose:
			ForceWindowRedraw(u->w);
			break;

		default:
			break;
		}

	}
	else if (XCheckTypedEvent(u->dpy, ClientMessage, &ev)) {

		if (ev.xclient.message_type == u->protocolsAtom
		&& (ev.xclient.data.l[0] == u->deleteWindowAtom
		|| ev.xclient.data.l[0] == u->closedownAtom)) {

			killPlayer(c, NULL);
			return TRUE;
		}

	}

	return FALSE;  /* done with all pending events for this viewer */

}


void
events_window_keyb_buttons(void)
/*
	Process events for each viewer
*/
{
	viewer   *u;
	_BOOL     got_evt;

#ifdef FIXME
	int n;

	n = 0;
	u = vl_head;
	while( u != NULL ){
		n++;
		u = u->vl_next;
	}
	printf("FIXME: trovati %d viewers\n", n);
#endif

	do {

		/*
			Process one event for each viewer
		*/

		got_evt = FALSE;

		u = vl_head;

		while( u != NULL ){

			if (real_delta_t || ((cur % redraw_interval) == 0)) {
				getStick(u->c, u);
			}

			if( XEventsQueued(u->dpy, QueuedAlready) ){

				got_evt = TRUE;

				if( doWindowEvent(u) ){

					/*
					Viewer or craft terminated. We can't
					continue scanning the linked list of
					viewers because we don't know how many
					viewers were attached to this craft, so
					we have to restart from the first
					viewer:
					*/

					u = vl_head;

				} else {

					u = u->vl_next;
				}

			} else {

				u = u->vl_next;
			}
		}
	
	} while( got_evt );
}


static int last_switches = 0;


int
events_joystick(craft * c, viewer * u, double throttle, int switches)
{

	int       switch_xor = switches ^ last_switches;

/*
 *  Change in state of any buttons ?
 */

	if (switch_xor != 0) {

		if (switch_xor & 1) {

/*
 *  Button 1 press
 */
			if (switches & 1) {
				fireWeapon(c);
			}

/*
 *  Button 1 release
 */

			else {
				ceaseFireWeapon(c);
			}
		}

/*
 * Button 2 press
 */

		if (switch_xor & 2) {
			if (switches & 2) {
				selectWeapon(c);
			}
		}

	}

	last_switches = switches;

	return 0;
}

