/*
 *  acm : an aerial combat simulator for X
 *  Copyright (C) 2007  Umberto Salsi
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 dated June, 1991.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program;  if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave., Cambridge, MA 02139, USA.
 */

#include <string.h>
#include "pm.h"
#include "patchlevel.h"
#include "damage.h"

#define EXPIRE_PERIOD 15.0
/* clear current msg after ... (s) */

#define HINT_PERIOD  30.0
/* if no msgs to show, show an hint after ... (s) */

static char *hints[] = {
"Hint: to exit the program press SHIFT-Q SHIFT-Q",
"Hint: to leave the window in background, press 'd'",
"Hint: check for updates at www.icosaedro.it/acm/download.html",
"Hint: press SHIFT-H to switch between HUD and classic instruments",
"Hint: press 'r' to show HSI, ADF and radar",
"Hint: one nautical mile corresponds to about 2 Km",
"Hint: knots is a speed unit corresponding to one nautical mile per hour",
"Hint: one feet is about one third of a meter",
"Hint: never exceed the maximum vertical load of the aircraft you are flying on!",
"Hint: your plane lacks the speed brakes? use the flaps instead!",
"Hint: press < and > for turns at standard rate at 3 DEG/s then press / or | to release",
"Hint: never exceed 25 DEG angle of bank for best passengers confort!",
"Hint: need a GPS? press SHIFT-D",
"Hint: don't forget to enable the rudder/aileron coordinator (SHIFT-X)",
"Hint: engine control keys: 1 (idle) 2 (decrease) 3 (increase) 4 (max)",
"Hint: to hold the current speed enable the auto-throttle (SHIFT-T)",
"Hint: to start/stop/reset the timer press 't'",
"Hint: to retract/deploy the landing gear press 'g'",
"Hint: to deploy the flaps press 'h' one or more times, retract with 'y'",
"Hint: to hold the current altitude, press SHIFT-Z",
"Hint: to hold the current climb rate press SHIFT-A",
"Hint: release the auto-pilot system with the Home key",
"Hint: select NAVAID frequency with 9 0, rotate the OBS pointer with 7 8",
"Hint: select RNAV waypoint with F3 F4 F5 F6",
"Hint: look around you: press NumLock then the arrows of the numeric keypad",
"Hint: switch between NAV/RNAV HSI receivers with SPACE",
"Hint: follow the VOR radial with the auto-navigator (SHIFT-N)",
"Hint: follow the ILS landing path with the auto-landing (SHIFT-L)",
"Hint: ACM-5.0-ico comes with a good manual: read it carefully!",
"Hint: set the zoom factor with + and -",
};

#define BUF_SIZE 100

typedef struct _prompt_data {
	struct _prompt_data *next;

	_BOOL have_msg;
	double expire;
	char buf[BUF_SIZE];

	double last_msg_shown_at;
} prompt_data;

static prompt_data *free_list = NULL;


void prompt_free(viewer *u)
{
	prompt_data *pd;

	if( u->prompt == NULL )
		return;
	
	pd = u->prompt;

	pd->next = free_list;
	free_list = pd;
	u->prompt = NULL;
}


void prompt_free_all()
{
	prompt_data *pd;

	while( free_list != NULL ){
		pd = free_list;
		free_list = pd->next;
		FREE(pd);
	}
}


void prompt_viewer_print(viewer *u, char *s)
{
	prompt_data *pd;

	if( u->prompt == NULL ){
		if( free_list == NULL ){
			pd = MALLOC( sizeof(prompt_data) );
		} else {
			pd = free_list;
			free_list = pd->next;
		}
		pd->next = NULL;
		pd->have_msg = FALSE;
		pd->expire = 0.0;
		pd->buf[0] = '\0';
		pd->last_msg_shown_at = -HINT_PERIOD; /* next msg shown immediately */
		u->prompt = pd;
	} else {
		pd = u->prompt;
	}

	pd->expire = curTime + EXPIRE_PERIOD;
	strncpy(pd->buf, s, BUF_SIZE);
	pd->buf[BUF_SIZE-1] = '\0';
	pd->have_msg = TRUE;
}


void prompt_craft_print(craft *c, char *s)
{
	viewer *v;

	v = vl_head;

	while( v != NULL ){
		if( v->c == c )
			prompt_viewer_print(v, s);
		v = v->next;
	}
}


void prompt_broadcast_print(char *s)
{
	viewer *v;

	v = vl_head;

	while( v != NULL ){
		prompt_viewer_print(v, s);
		v = v->next;
	}
}


static void draw_string(Viewport *v, prompt_data *pd)
{
	double fh, fw, w, k;
	int width, height, l, x, y;

	width = RectWidth(v->rect);
	height = RectHeight(v->rect);
	fh = ceil(height / 30.0);
	fw = VFontWidthPixels(v, (int) (fh+0.5));
	l = strlen(pd->buf);
	k = l * fw / (0.95*width);
	if( k >= 1.0 ){
		fh = floor(fh/k);
		fw = VFontWidthPixels(v, (int) (fh+0.5));
	}
	w = fw * l;

	AlibSetClipRect(v->w, &v->rect);
	x = v->rect.a.x + width/2 - w/2.0;
	y = v->rect.b.y - fh;
	VDrawStrokeString(v, x + 1, y,
		pd->buf, l, fh, v->w->pixel[blackColor->cIndex]);
	VDrawStrokeString(v, x, y,
		pd->buf, l, fh, v->w->pixel[whiteColor->cIndex]);
}


void prompt_draw(viewer *u)
{
	prompt_data *pd;
	int n, i;

	if( u->prompt == NULL ){
		prompt_viewer_print(u, "**** Wellcome to ACM-"
			REVISION_STRING " ****");
		return;
	}
	
	pd = u->prompt;

	/*
		Display hints only in stright, level, at constant speed flight
	*/
	if(
		fabs(u->c->Cg.z)*60.0 > 90.0 /* ft/min */
	|| fabs(u->c->curRoll) > DEGtoRAD(5.0)
	|| fabs(u->c->G.x) > 0.1
	)
		pd->last_msg_shown_at = curTime + HINT_PERIOD;


	if( pd->have_msg ){
		/*
			Show current msg:
		*/

		if( curTime > pd->expire ){
			pd->have_msg = FALSE;
			pd->last_msg_shown_at = curTime;
			return;
		}

		draw_string(u->v, pd);
	
	} else if( pd->last_msg_shown_at + HINT_PERIOD < curTime ){
		/*
			Show hint:
		*/

		n = sizeof(hints) / sizeof(char *);
		i = damage_random() % n;

		prompt_viewer_print(u, hints[i]);
	}
}

/* End of the prompt module. */
